import { createWriteStream } from 'fs';
import { readFile, unlink } from 'fs/promises';

interface CacheableResponse {
	statusCode: number;
	statusMessage?: string;
	contentType?: string;
	data?: unknown;
}

const getFilePathFromRequestId = (requestId: string) => `${requestId}`;

export const cacheResponse = async (requestId: string, res: CacheableResponse) => {
	const filePath = getFilePathFromRequestId(requestId);
	const cacheStream = createWriteStream(filePath, { encoding: 'utf-8' });
	cacheStream.write(`${res.statusCode.toString()} ${res.statusMessage || ''}\n`);
	if (res.contentType) {
		cacheStream.write(`Content-Type: ${res.contentType}\n`);
	}
	if (res.data) {
		cacheStream.write('\n');
		cacheStream.write(res.data as string);
	}

	return new Promise((resolve) => {
		cacheStream.close(resolve);
	});
};

export const retrieveCache = async (requestId: string) => {
	const filePath = getFilePathFromRequestId(requestId);
	const requestBuffer = await readFile(filePath, 'utf-8');
	await unlink(filePath);
	const [statusLine, ...headersAndBody] = requestBuffer.split('\n');
	const [statusCode, ...statusMessageWords] = statusLine.split(' ');
	const statusMessage = statusMessageWords.join(' ');
	const bodyStart = headersAndBody.findIndex((line) => line === '');
	const headers = headersAndBody.slice(0, bodyStart);
	const body = headersAndBody.slice(bodyStart + 1).join('\n');
	const contentTypeHeader = headers.find((header) => header.toLowerCase().startsWith('content-type:'));
	const contentType = contentTypeHeader?.split(':')[1].trim();
	return {
		statusCode: parseInt(statusCode, 10),
		statusMessage,
		contentType,
		body,
	};
};
