import { ParsedUrlQuery } from 'querystring';
import { GetServerSideProps, GetServerSidePropsContext, PreviewData } from 'next';
import { deserialize } from 'seroval';
import { NextApiRequest, NextApiResponse } from '../common/types';
import { DEFAULT_ENCODING, DEFAULT_METHOD } from './constants';
import { getBody } from '../utils/request';
import { REQUEST_ID_COOKIE_KEY, CookieManager } from '../utils/cookies';
import { ENCTYPE_APPLICATION_JSON, ENCTYPE_APPLICATION_OCTET_STREAM } from '../common/enctypes';
import { METHODS_WITH_BODY } from '../common/constants';
import { retrieveCache } from './cache';

export type DestinationGetServerSideProps<
	Props extends Record<string, unknown> = Record<string, unknown>,
	Params extends ParsedUrlQuery = ParsedUrlQuery,
	Preview extends PreviewData = PreviewData,
> = (
	actionReq: NextApiRequest,
	actionRes: NextApiResponse,
	context: GetServerSidePropsContext<Params, Preview>,
) => ReturnType<GetServerSideProps<Props, Params, Preview>>;

export interface DestinationWrapperOptions {
	fn?: DestinationGetServerSideProps;
}

export const getServerSideProps = (
	options = {} as DestinationWrapperOptions,
): GetServerSideProps => async (ctx) => {
	const req: NextApiRequest = {
		query: {
			...ctx.query,
			...(ctx.params ?? {}),
		},
		body: null,
	};
	const { method = DEFAULT_METHOD } = ctx.req;

	if (METHODS_WITH_BODY.includes(method.toUpperCase() as typeof METHODS_WITH_BODY[number])) {
		const body = await getBody(ctx.req);
		req.body = body.toString(DEFAULT_ENCODING);
	}

	const res: NextApiResponse = {};

	const cookieManager = new CookieManager(ctx);
	const resRequestId = cookieManager.getCookie(REQUEST_ID_COOKIE_KEY);
	cookieManager.unsetCookie(REQUEST_ID_COOKIE_KEY);
	if (resRequestId) {
		const {
			statusCode,
			statusMessage,
			contentType,
			body: resBody,
		} = await retrieveCache(resRequestId);
		ctx.res.statusCode = statusCode;
		ctx.res.statusMessage = statusMessage;
		if (contentType === ENCTYPE_APPLICATION_JSON) {
			res.body = deserialize(resBody);
		} else if (contentType === ENCTYPE_APPLICATION_OCTET_STREAM) {
			res.body = deserialize(resBody);
		} else {
			const c = console;
			c.warn('Could not parse response body, returning nothing');
			res.body = null;
		}
	}

	const gspResult = (
		typeof options?.fn === 'function'
			? await options.fn(req, res, ctx)
			: {
				props: {},
			}
	);

	if ('props' in gspResult) {
		return {
			...gspResult,
			props: {
				...gspResult.props,
				req,
				res,
			},
		};
	}

	// redirect/not found will be treated as default behavior
	return gspResult;
};
